"""
A class to generate a Serotype

All Serotyping logic stays here.
"""

import os
import json
import datetime
import getpass
import hashlib
import shlex
import copy
import pkg_resources

from loguru import logger

from .Blast import MakeBlastDB


SEROTYPE_FASTA = "db/sequences.fasta"


def report_maker(full_matches, partial_matches=[]):
    report = {
        "prs": None,
        "lmo0737": None,
        "lmo1118": None,
        "orf2110": None,
        "orf2819": None,
    }
    for gene in report:
        if gene.upper() in full_matches:
            report[gene] = "FULL"
        elif gene.upper() in partial_matches:
            report[gene] = "PARTIAL"
        else:
            report[gene] = "NONE"
    if "PRS" not in full_matches:
        report["serotype"] = "Nontypeable"
        report["comment"] = "No Prs found"
    elif (
        "LMO0737" in full_matches
        and "ORF2819" not in full_matches
        and "ORF2110" not in full_matches
    ):
        if "LMO1118" in full_matches:
            report["serotype"] = "1/2c, 3c"
            report["comment"] = None
        else:
            report["serotype"] = "1/2a, 3a"
            report["comment"] = None
    elif (
        "ORF2819" in full_matches
        and "LMO0737" not in full_matches
        and "LMO1118" not in full_matches
    ):
        if "ORF2110" in full_matches:
            report["serotype"] = "4b, 4d, 4e"
            report["comment"] = None
        else:
            report["serotype"] = "1/2b, 3b, 7"
            report["comment"] = None
    elif (
        "LMO0737" in full_matches
        and "ORF2819" in full_matches
        and "ORF2110" in full_matches
    ):
        if "LMO1118" in full_matches:
            report["serotype"] = "Nontypeable"
            report["comment"] = "Presence of all 5 genes, not a known serotype"
        else:
            report["serotype"] = "4b, 4d, 4e*"
            report["comment"] = "Unusual 4b with lmo0737"
    else:
        report["serotype"] = "Nontypeable"
        report["comment"] = "No combination of fully matched genes resulted in a known serotype"
    logger.debug(report)
    return report


class Typing:

    BLAST_OUTFMT = "6"
    OUTPUT_HEADER = ["ID", "TYPE", "DB_VERSION"]

    def __init__(self, blast_run, db, cov=100, pid=100):
        self.db_version = db.version
        path_db = os.path.realpath(db.db_name)
        self.blast = copy.deepcopy(blast_run)
        self.blast.add_db(path_db)
        self.blast.add_option("-ungapped")
        self.blast.add_option("-culling_limit", 1)
        self.blast.add_option("-outfmt", self.BLAST_OUTFMT)
        self.blast.add_option("-dust", "no")
        self.cov = cov
        self.pid = pid

    def __str__(self):
        try:
            string = json.dumps(self.report, indent=4)
        except:
            string = "{}"
        return string

    def _blast_run(self):
        self.blast_res = self.blast.run()

    def _blast_parse(self):
        self.full_matches = set()
        self.partial_matches = set()
        logger.info(self.blast_res.stdout.strip())
        blast_matches = self.blast_res.stdout.strip().split("\n")
        for b in blast_matches:
            (qaccver, saccver, length, slen, pident) = b.split("\t")
            if (
                float(pident) >= self.pid
                and 100 * (float(length) / float(slen)) >= self.cov
            ):
                self.full_matches.update([saccver.split("~~")[0]])
            else:
                self.partial_matches.update([saccver.split("~~")[0]])

    def tab_report(self, header=True, sep=","):
        if header:
            print(sep.join(self.OUTPUT_HEADER))
        out = []
        for field in self.OUTPUT_HEADER:
            out += [self.report[field.lower()]]
        print(sep.join(out))

    def csv_report(self, header=True):
        self.tab_report(sep=",", header=header)

    def tsv_report(self, header=True):
        self.tab_report(sep="\t", header=header)


class Serotype(Typing):
    """
    Serotyping of Listeria monocytogenes samples is performed throught the
    presence abscense of five genes.

    One gene must be present to ascert that the species is indeed Listeria
    monocytogenes (Prs). If the gene is not present, then Serotyping CANNOT
    proceed.

    Serogroup 1/2a, 3a requires the additional presence of: lmo0737

    Serogroup 1/2b, 3b, 7 requires the additional presence of: ORF2819

    Serogroup 1/2c, 3c requires the additional presence of: lmo1118 & lmo0737

    Serogroup 4b, 4d, 4e requires the additional presence of: ORF2110 & ORF2819

    Serogroup 4b, 4d, 4e* requires the additional presence of: ORF2110 & ORF2819 & lmo0737

    This creates a decision tree.

    What are the possible outputs?
        Alleles in {'', Prs, lmo0737, lmo1118, ORF2819, ORF2110}

    if not Prs:
        stop
    elif lmo0737 and not (ORF2819 or ORF2110):
        if lmo1118:
            Serogroup 1/2c, 3c
        else:
            Serogroup 1/2a, 3a
    elif ORF2819 and not (lmo0737 or lmo1118):
        if ORF2110:
            Serogroup 4b, 4d, 4e
        else:
            Serogroup 1/2b, 3b, 7
    elif lmo0737 and ORF2819 and ORF2110:
        Serogroup 4b, 4d, 4e*
    else:
        Nontypable
    """

    BLAST_OUTFMT = "6 qaccver saccver length slen pident"
    OUTPUT_HEADER = [
        "ID",
        "SEROTYPE",
        "Prs",
        "lmo0737",
        "lmo1118",
        "ORF2110",
        "ORF2819",
        "COMMENT",
        "DB_VERSION",
    ]

    def __init__(self, blast_run, db, cov=100, pid=98):
        super().__init__(blast_run=blast_run, db=db, cov=cov, pid=pid)

    def _blast_parse(self):
        self.full_matches = set()
        self.partial_matches = set()
        blast_matches = self.blast_res.stdout.strip().split("\n")
        for b in blast_matches:
            (qaccver, saccver, length, slen, pident) = b.split("\t")
            obs_pid = float(pident)
            obs_cov = 100 * (float(length) / float(slen))
            if obs_pid >= self.pid and obs_cov >= self.cov:
                self.full_matches.update([saccver.split("~~")[0].upper()])
            else:
                self.partial_matches.update([saccver.split("~~")[0].upper()])

    def generate_type(self, query):
        """
        report = {
            "prs": None,
            "lmo0737": None,
            "lmo1118": None,
            "orf2110": None,
            "orf2819": None,
        }
        """
        self.blast.add_query(query)
        self._blast_run()
        self._blast_parse()
    #Change function structure to implement pytest
        report = report_maker(self.full_matches, self.partial_matches)
        report["id"] = query
        report["db_version"] = self.db_version()
        self.report = report


class SerotypeDB:
    def __init__(
        self,
        path_db,
        db_type,
        db_name="lissero",
        title="Listeria Serotyping BLAST DB",
        makeblastdb_path=None,
        force=False,
    ):
        self.title = title
        self.path_db = os.path.realpath(path_db)
        self.db_name = os.path.join(self.path_db, db_name)
        self.mkdb = MakeBlastDB(makeblastdb_path=makeblastdb_path)
        if db_type == "serotype":
            self.infile = pkg_resources.resource_filename("lissero", SEROTYPE_FASTA)
        else:
            logger.critical(f"I don't understand db_type = {db_type}")
            raise IOError
        self.force = force
        self.log_file = os.path.join(self.path_db, db_name + "_db.json")
        self.db_log = {}

    def check_db(self):
        if self.force and self.infile is not None:
            logger.info("Will rebuild database if one exists!")
            self._make_db()
            return
        elif os.path.exists(self.db_name + ".nhr"):
            self._load_log()
            logger.info(f"DB was created on {self.db_log['date_created']}")
            logger.info("DB is ready for use.")
        elif self.infile is None:
            logger.critical(
                "There is no database, and I don't"
                " have enough information to create one."
                " Please add an input file for the DB"
                " and run again."
            )
            raise IOError
        else:
            logger.warning(f"Did not find a DB at {self.path_db}")
            logger.warning("Will create a new one")
            self._make_db()

    def __str__(self):
        return self.db_name

    def version(self):
        return self.db_log["date_created"]

    def _load_log(self):
        if os.path.exists(self.log_file):
            fh = open(self.log_file, "rt")
            self.db_log = json.load(fh)
            logger.info("Successfully loaded DB log!")
            fh.close()
        else:
            logger.warning("Did find a log file at {self.path_db}")

    def _save_log(self):
        fh = open(self.log_file, "wt")
        json.dump(self.db_log, fh, indent=4)
        logger.info("Successfully saved log!")

    def _hash_file(self, filename):
        f = open(filename, "rb")
        file_content = f.read()
        h = hashlib.sha256(file_content).hexdigest()
        f.close()
        return h

    def _make_db(self):
        self.db_log["date_created"] = datetime.datetime.today().isoformat()
        mkdb = self.mkdb
        mkdb.add_option("-dbtype", "nucl")
        mkdb.add_option("-in", self.infile)
        mkdb.add_option("-parse_seqids")
        mkdb.add_option("-input_type", "fasta")
        mkdb.add_option("-out", self.db_name)
        mkdb.add_option("-title", shlex.quote(self.title))
        mkdb.run()
        self.db_log["cmd"] = str(mkdb)
        self.db_log["creator"] = getpass.getuser()
        self.db_log["infile"] = {}
        self.db_log["infile"]["path"] = self.infile
        self.db_log["infile"]["sha256"] = self._hash_file(self.infile)
        self.db_log["title"] = self.title
        self._save_log()
        logger.info(f"Successfully created {self.db_name}")



"""
A sample class for LisSero
"""

import os

from loguru import logger

from .Serotype import Serotype


class Sample:
    def __init__(
        self,
        filename,
        blast,
        sero_db,
        sg_min_id,
        sg_min_cov,
    ):
        self.id = filename
        self.filename = os.path.realpath(filename)
        self.serotype = Serotype(blast, sero_db, pid=sg_min_id, cov=sg_min_cov)

    def is_fasta(self):
        pass

    def get_serotype(self):
        logger.info(f"Serotyping: {self.id}")
        self.serotype.generate_type(self.filename)

    def __str__(self):
        try:
            string = f"{self.id} " \
                     f"\tSEROTYPE: {self.serotype.report['serotype']}"
        except:
            string = ""
        return string


class Samples:

    SIMPLE_HEADER = ["ID", "SEROTYPE", "PRS", "LMO0737", "LMO1118", "ORF2110", "ORF2819", "COMMENT"]

    def __init__(
        self,
        filenames,
        blast,
        sero_db,
        sg_min_id,
        sg_min_cov,
    ):
        self.filenames = filenames
        self.blast = blast
        self.sero_db = sero_db
        self.sg_min_id = sg_min_id
        self.sg_min_cov = sg_min_cov

    def _create_sample(self):
        self.samples = []
        for f in self.filenames:
            self.samples += [
                Sample(
                    f,
                    blast=self.blast,
                    sero_db=self.sero_db,
                    sg_min_id=self.sg_min_id,
                    sg_min_cov=self.sg_min_cov,
                )
            ]

    def _run_typing(self, func):
        if func == "serotype":
            for f in self.samples:
                f.get_serotype()
        else:
            logger.critical(f"Unknown function {func}")
            raise RuntimeError

    def run_typing(self):
        self._create_sample()
        self._run_typing("serotype")

    def simple_report(self, header=True):
        print("\t".join(self.SIMPLE_HEADER))
        for sample in self.samples:
            sample_id = sample.serotype.report["id"]
            serotype = sample.serotype.report["serotype"]
            prs = sample.serotype.report["prs"]
            lmo0737 = sample.serotype.report["lmo0737"]
            lmo1118 = sample.serotype.report["lmo1118"]
            orf2110 = sample.serotype.report["orf2110"]
            orf2819 = sample.serotype.report["orf2819"]
            comment = sample.serotype.report["comment"]
            if comment is None:
                comment = ""
            print("\t".join([sample_id, serotype, prs, lmo0737, lmo1118, orf2110, orf2819, comment]))

    def __str__(self):
        return f"Typing {len(self.filenames)} samples"



"""
A class to deal with Blasting
"""

import subprocess
import shutil
import os
import re

from loguru import logger


class SubRunner:

    version_pat = re.compile(r"(\d+)\.(\d+)\.(\d+)")

    def __init__(self, cmd, cmd_path=None):
        self.cmd = cmd
        if cmd_path is not None:
            self.cmd_path = os.path.realpath(cmd_path)
        else:
            self.cmd_path = shutil.which(self.cmd)
            if self.cmd_path is None:
                logger.critical(f"Could not find executable {self.cmd}")
                logger.critical("Please provide a path to the executable.")
                raise SystemError
        self.cmd_list = [self.cmd_path]

    def add_option(self, key, value=None):
        self.cmd_list += [str(key)]
        if value is not None:
            self.cmd_list += [str(value)]

    def version(self):
        self.add_option("-version")
        res = self.run()
        try:
            self.version_no = self.version_pat.findall(res.stdout)[0]
            major, minor, patch = self.version_no
        except:
            logger.critical(
                f"Something went wrong when parsing the version" f" for {self.cmd}"
            )
            raise SystemError
        logger.info(f"Found {self.cmd} version {major}.{minor}.{patch}")
        self.cmd_list = [self.cmd_path]

    def is_version(self, requirement):
        pass

    def run(self, stdout=subprocess.PIPE, stderr=subprocess.PIPE, retcode=0):
        logger.info("Blastn is running: " + ' '.join(self.cmd_list))
        p = subprocess.run(self.cmd_list, stdout=stdout, stderr=stderr, encoding="utf8")
        if p.returncode != retcode:
            logger.critical(f"Failed to run {' '.join(self.cmd_list)}")
            raise SyntaxError
        return p

    def __str__(self):
        return " ".join(self.cmd_list)


class Blast(SubRunner):
    def __init__(self, blast_path=None):
        super().__init__(cmd="blastn", cmd_path=blast_path)
        pass

    def add_db(self, path):
        path = os.path.realpath(path)
        self.add_option("-db", path)

    def add_query(self, path):
        path = os.path.realpath(path)
        self.add_option("-query", path)


class MakeBlastDB(SubRunner):
    def __init__(self, makeblastdb_path=None):
        super().__init__(cmd="makeblastdb", cmd_path=makeblastdb_path)
        pass



#!/usr/bin/env python
# Script by Jason Kwong
# In silico serotyping for L.monocytogenes
import sys
import logging
import os

import click
import loguru
from Bio import SeqIO

from lissero.scripts.Sample import Samples
from lissero.scripts.Blast import Blast
from lissero.scripts.Serotype import SerotypeDB
from .__init__ import __version__ as version

logger = loguru.logger


def print_version(ctx, param, value):
    if not value or ctx.resilient_parsing:
        return
    click.echo(f"LisSero {version}")
    ctx.exit()


def is_fasta(filename):
    """
    There are no real FASTA validators out there.
    This is the best I could come up with. If the
    file is empty or does not contain any FASTA records
    the parser will return an empty generator which will
    return false in `any(gen)`.
    But, there is another case where the file starts with
    a `>`, and thus the generator works, but it returns an
    empty record, thus the `and len(next(gen)) > 0.

    Args:
        filename: FASTA input name

    Returns:
        boolean: true if it looks like a FASTA false otherwise.
    """
    gen = SeqIO.parse(filename, "fasta")
    try:
        rec = next(gen)
        return len(rec) > 0
    except StopIteration:
        return False
    except Exception as e:
        logger.error(e)
        sys.exit(1)

@click.command()
@click.option("-s", "--serotype_db", default=None, envvar="LISSERO_DB")
@click.option(
    "--min_id",
    default=95.0,
    help="Minimum percent identity to accept a match. [0-100]",
    show_default=True,
)
@click.option(
    "--min_cov",
    default=95.0,
    help="Minimum coverage of the gene to accept a match. [0-100]",
    show_default=True,
)
@click.option("--debug", is_flag=True)
@click.option("--logfile", help="Save log to a file instead of printing to stderr", default="")
@click.argument("fasta", nargs=-1, type=click.Path(exists=True), required=True)
# fix Version Issue #10
@click.option("--version", is_flag=True, callback=print_version,
              expose_value=False, is_eager=True, help="Show Version Information")
def run_lissero(serotype_db, min_id, min_cov, debug, logfile, fasta):
    """
    In silico serogroup prediction for L. monocytogenes.
    Alleles: lmo1118, lmo0737, ORF2819, ORF2110, Prs

    References:

    * Doumith et al. Differentiation of the major Listeria monocytogenes
        serovars by multiplex PCR. J Clin Microbiol, 2004; 42:8; 3819-22

    """

    if debug:
        log_level = logging.DEBUG
    else:
        log_level = logging.INFO

    if logfile != "":
        logger.remove()
        logger.add(logfile, level=log_level)
    else:
        logger.remove()
        logger.add(sys.stderr, level=log_level)

    all_fasta = all([is_fasta(fna) for fna in fasta])
    if not all_fasta:
        logger.error("One or more input files do not appear to be "
                     "valid FASTA.")
        sys.exit(1)

    # fix issue #11 and #12
    path_serodb = ""
    try:
        path_serodb = os.path.realpath(serotype_db)
    except TypeError as e:
        logger.error(f"Please provide a correct serotype db path or set correct PATH for LISSERO_DB")
        sys.exit(1)
    sero_db = SerotypeDB(path_db=path_serodb, db_type="serotype")
    sero_db.check_db()
    blast = Blast()
    samples = Samples(
        fasta, blast=blast, sero_db=sero_db, sg_min_id=min_id, sg_min_cov=min_cov
    )
    samples.run_typing()
    samples.simple_report()


if __name__ == "__main__":
    run_lissero()
